#include <gecode/driver.hh>
#include <gecode/int.hh>
#include <gecode/float.hh>
#include <gecode/set.hh>
#include <gecode/minimodel.hh>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <time.h>
#include <sys/resource.h>
#include <vector>
#include <math.h>
#include <algorithm> 


using namespace Gecode;
using namespace std;




/*
	Class of branching for the criterion of Diamter/Split
	Using the matrix of distance: dd[i][j]
	Idea: 
		- choice the first point that has Min Size of Domain
		- Assign this point to the closest cluster 
		(distance between point i and a cluster is defined as the maximum distance betwwen i and points in the cluster)
*/
class branchDiameter : public Brancher {
protected:
  /// Views of the brancher
  ViewArray<Int::IntView> x; // x is the variables G

  /// Next variable to branch on
  mutable int start;	// start is the first point that is unassigned

  /// %Choice
  class Choice : public Gecode::Choice {
  public:
    /// Position of variable
    int pos;
    /// Value of variable
    int val;
    /** Initialize choice for brancher \a b, position \a pos0, 
     *  and value \a val0.
     */
    Choice(const Brancher& b, int pos0, int val0)
      : Gecode::Choice(b,2), pos(pos0), val(val0) {}

    /// Report size occupied
    virtual size_t size(void) const {
      return sizeof(Choice);
    }

    /// Archive into \a e
    virtual void archive(Archive& e) const {
      Gecode::Choice::archive(e);
      e << pos << val;
    }
  };
 
  /// Construct brancher
  branchDiameter(Home home, ViewArray<Int::IntView>& xv) 
    : Brancher(home), x(xv), start(0) {}
  /// Copy constructor
  branchDiameter(Space& home, bool share, branchDiameter& b) 
    : Brancher(home, share, b), start(b.start) {
    x.update(home, share, b.x);
  }

public:
  /// Check status of brancher, return true if alternatives left
  virtual bool status(const Space&) const {
    for (int i=start; i<x.size(); i++)
      if (!x[i].assigned()) {
        start = i; return true;
      }
    return false;
  }

  /// Return choice
  virtual Gecode::Choice* choice(Space&) {
    int n = x.size(); // there is n point
    unsigned int minSize = 100000, i = 0;
    
    for (int j = start; j < n; j++)  	// search all point that are not assigned (not instantiated)
      if ((!x[j].assigned()) && (minSize > x[j].size())) {   // choice the first point that has Min Size of Domain
				minSize = x[j].size();
				i = j;
      }
    // now What valeur of x[i]
    int minVal = 0; double minDist = 10000000;
	
	double check[k];

	for (int t = 0; t < k; t++)
		check[t] = 0;

	for (int t = 0; t < n; t++)
		if (x[t].assigned())
				if (x[i].in(x[t].val())) 
					if (dd[i][t] > check[x[t].val()]) 
						check[x[t].val()] = dd[i][t];
				
	for (int c = 0; c < k; c++)
		if (x[i].in(c)) 
			if (minDist > check[c]) {
				minDist = check[c];
				minVal = c;
			}

    return new Choice(*this,i,minVal);		// assign G[i] = minVal, G[i] != minVal
  }

  /// Return choice
  virtual Choice* choice(const Space&, Archive& e) {
    int pos, val;
    e >> pos >> val;
    return new Choice(*this, pos, val);
  }

  /// Perform commit for choice \a _c and alternative \a a
  virtual ExecStatus commit(Space& home, const Gecode::Choice& _c, unsigned int a) {
    const Choice& c = static_cast<const Choice&>(_c);
    if (a == 0)
      return me_failed(x[c.pos].eq(home, c.val)) ? ES_FAILED : ES_OK;
    else 
      return me_failed(x[c.pos].nq(home, c.val)) ? ES_FAILED : ES_OK;
  }

  /// Copy brancher
  virtual Actor* copy(Space& home, bool share) {
    return new (home) branchDiameter(home, share, *this);
  }

  /// Post brancher
  static void post(Home home, const IntVarArgs& x) {
    ViewArray<Int::IntView> xv(home, x);
    (void) new (home) branchDiameter(home, xv);
  }

  /// Delete brancher and return its size
  virtual size_t dispose(Space&) {
    return sizeof(*this);
  }
};
